# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the core API functionality for interacting with the Ansys HPS data transfer client.

This module includes methods and utilities for performing
data transfer operations, managing resources, and handling client interactions.
"""

import builtins
from collections.abc import Callable
import logging
import time
import traceback

import backoff

from ..client import Client
from ..exceptions import TimeoutError
from ..models import (
    CheckPermissionsResponse,
    DataAssignment,
    GetPermissionsResponse,
    Operation,
    OperationIdResponse,
    OperationsResponse,
    OperationState,
    RoleAssignment,
    RoleQuery,
    SetMetadataRequest,
    SrcDst,
    StatusResponse,
    StorageConfigResponse,
    StoragePath,
)
from ..utils.jitter import get_expo_backoff
from .handler import WaitHandler
from .retry import retry

log = logging.getLogger(__name__)


class DataTransferApi:
    """Provides the data transfer API.

    Parameters
    ----------
    client: Client
        Client object.
    """

    def __init__(self, client: Client):
        """Initializes the DataTransferApi class object."""
        self.dump_mode = "json"
        self.client = client
        self.wait_handler_factory = WaitHandler

    @retry()
    def status(self, wait=False, sleep=5, jitter=True, timeout: float | None = 20.0):
        """Get the status of the worker binary."""

        def _sleep():
            log.info(f"Waiting for the worker to be ready on port {self.client.binary_config.port} ...")
            s = backoff.full_jitter(sleep) if jitter else sleep
            time.sleep(s)

        url = "/"
        start = time.time()
        while True:
            if timeout is not None and (time.time() - start) > timeout:
                raise TimeoutError("Timeout waiting for worker to be ready")

            resp = self.client.session.get(url)
            json = resp.json()
            s = StatusResponse(**json)
            if wait and not s.ready:
                _sleep()
                continue
            return s

    @retry()
    def operations(self, ids: list[str], expand: bool = False):
        """Get a list of operations.

        Parameters
        ----------
        ids: List[str]
            List of IDs.
        """
        return self._operations(ids, expand=expand)

    def storages(self):
        """Get types of storages available on the storage backend."""
        url = "/storage"
        resp = self.client.session.get(url)
        json = resp.json()
        return StorageConfigResponse(**json).storage

    def copy(self, operations: list[SrcDst]):
        """Get the API response for copying a list of files.

        Parameters
        ----------
        operations: List[SrcDst]
        """
        return self._exec_operation_req("copy", operations)

    def exists(self, operations: list[StoragePath]):
        """Check if a path exists.

        Parameters
        ----------
        operations: List[StoragePath]
        """
        return self._exec_operation_req("exists", operations)

    def list(self, operations: list[StoragePath]):
        """List files in a path.

        Parameters
        ----------
        operations: List[StoragePath]
        """
        return self._exec_operation_req("list", operations)

    def mkdir(self, operations: builtins.list[StoragePath]):
        """Create a directory.

        Parameters
        ----------
        operations: List[StoragePath]
        """
        return self._exec_operation_req("mkdir", operations)

    def move(self, operations: builtins.list[SrcDst]):
        """Move a file on the backend storage.

        Parameters
        ----------
        operations: List[SrcDst]
        """
        return self._exec_operation_req("move", operations)

    def remove(self, operations: builtins.list[StoragePath]):
        """Delete a file.

        Parameters
        ----------
        operations: List[StoragePath]
        """
        return self._exec_operation_req("remove", operations)

    def rmdir(self, operations: builtins.list[StoragePath]):
        """Delete a directory.

        Parameters
        ----------
        operations: List[StoragePath]
        """
        return self._exec_operation_req("rmdir", operations)

    @retry()
    def _exec_operation_req(
        self, storage_operation: str, operations: builtins.list[StoragePath] | builtins.list[SrcDst]
    ):
        url = f"/storage:{storage_operation}"
        payload = {"operations": [operation.model_dump(mode=self.dump_mode) for operation in operations]}
        resp = self.client.session.post(url, json=payload)
        json = resp.json()
        r = OperationIdResponse(**json)
        return r

    def _operations(self, ids: builtins.list[str], expand: bool = False):
        url = "/operations"
        params = {"ids": ids}
        if expand:
            params["expand"] = "true"
        resp = self.client.session.get(url, params=params, timeout=1)
        json = resp.json()
        return OperationsResponse(**json).operations

    @retry()
    def check_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Check permissions of a path (including parent directory) using a list of ``RoleAssignment`` objects.

        Parameters
        ----------
        permissions: List[RoleAssignment]
        """
        url = "/permissions:check"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        resp = self.client.session.post(url, json=payload)
        json = resp.json()
        return CheckPermissionsResponse(**json)

    @retry()
    def get_permissions(self, permissions: builtins.list[RoleQuery]):
        """Get permissions of a file from a list of ``RoleQuery`` objects.

        Parameters
        ----------
        permissions: List[RoleQuery]
        """
        url = "/permissions:get"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        resp = self.client.session.post(url, json=payload)
        json = resp.json()
        return GetPermissionsResponse(**json)

    @retry()
    def remove_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Remove permissions using a list of ``RoleAssignment`` objects.

        Parameters
        ----------
        permissions: List[RoleAssignment]
        """
        url = "/permissions:remove"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        self.client.session.post(url, json=payload)

    @retry()
    def set_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Set permissions using a list of ``RoleAssignment`` objects.

        Parameters
        ----------
        permissions: List[RoleAssignment]
        """
        url = "/permissions:set"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        self.client.session.post(url, json=payload)

    @retry()
    def get_metadata(self, paths: builtins.list[str | StoragePath]):
        """Get metadata of a path on the backend storage.

        Parameters
        ----------
        paths: List[str | StoragePath]
        """
        url = "/metadata:get"
        paths = [p if isinstance(p, str) else p.path for p in paths]
        payload = {"paths": paths}
        resp = self.client.session.post(url, json=payload)
        json = resp.json()
        return OperationIdResponse(**json)

    @retry()
    def set_metadata(self, asgs: dict[str | StoragePath, DataAssignment]):
        """Set metadata for a path on the backend storage.

        Parameters
        ----------
        asgs: Dict[str | StoragePath, DataAssignment]
            List of paths with key of type string or ``StoragePath`` and value of ``DataAssignment``.
        """
        url = "/metadata:set"
        d = {k if isinstance(k, str) else k.path: v for k, v in asgs.items()}
        req = SetMetadataRequest(metadata=d)
        resp = self.client.session.post(url, json=req.model_dump(mode=self.dump_mode))
        json = resp.json()
        return OperationIdResponse(**json)

    def wait_for(
        self,
        operation_ids: builtins.list[str | Operation | OperationIdResponse],
        timeout: float | None = None,
        interval: float = 0.1,
        cap: float = 2.0,
        raise_on_error: bool = False,
        handler: Callable[[builtins.list[Operation]], None] = None,
    ):
        """Wait for operations to complete.

        Parameters
        ----------
        operation_ids: List[str | Operation | OperationIdResponse]
            List of operation ids.
        timeout: float | None
            Timeout in seconds. Default is None.
        interval: float
            Interval in seconds. Default is 0.1.
        cap: float
            The maximum backoff value used to calculate the next wait time. Default is 2.0.
        raise_on_error: bool
            Raise an exception if an error occurs. Default is False.
        operation_handler: Callable[[builtins.list[Operation]], None]
            A callable that will be called with the list of operations when they are fetched.
        """
        if handler is None:
            handler = self.wait_handler_factory()

        if not isinstance(operation_ids, list):
            operation_ids = [operation_ids]
        operation_ids = [op.id if isinstance(op, Operation | OperationIdResponse) else op for op in operation_ids]
        start = time.time()
        attempt = 0
        while True:
            attempt += 1
            try:
                expand = getattr(handler.Meta, "expand_group", False) if hasattr(handler, "Meta") else False
                ops = self._operations(operation_ids, expand=expand)
                if handler is not None:
                    try:
                        handler(ops)
                    except Exception as e:
                        log.warning(f"Handler error: {e}")
                        log.debug(traceback.format_exc())

                if all(op.state in [OperationState.Succeeded, OperationState.Failed] for op in ops):
                    log.debug("All operations have completed.")
                    break
            except Exception as e:
                log.debug(f"Error getting operations: {e}")
                if raise_on_error:
                    raise
                else:
                    break

            if timeout is not None and (time.time() - start) > timeout:
                raise TimeoutError("Timeout waiting for operations to complete")

            # TODO: Adjust based on transfer speed and file size
            duration = get_expo_backoff(interval, attempts=attempt, cap=cap, jitter=True)
            time.sleep(duration)

        return ops



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for configuring OpenTelemetry (Otel) settings for the Ansys HPS data transfer client.

This module allows setting up environment variables required
for telemetry data export and resource attributes.
"""

import os


def set_otel_config(exporter_url, resource_attributes=None, headers=None, exporter_type=None):
    """Set data transfer worker Otel configuration using environment variables before starting the data transfer worker.

    ANSYS_DT_OTEL__EXPORTER_URL - Otel exporter url.
    ANSYS_DT_OTEL__RESOURCE_ATTRIBUTES - key-value pairs of resource attributes to be passed to the Otel SDK.
    ANSYS_DT_OTEL__HEADERS - key-value pairs of headers to be associated with gRPC requests.
    ANSYS_DT_OTEL__EXPORTER_TYPE - Otel exporter type.
    ANSYS_DT_OTEL__ENABLED - enables Otel.
    """
    os.environ["ANSYS_DT_OTEL__ENABLED"] = "True"
    if exporter_url is not None:
        os.environ["ANSYS_DT_OTEL__EXPORTER_URL"] = str(exporter_url)
    if exporter_type is not None:
        os.environ["ANSYS_DT_OTEL__EXPORTER_TYPE"] = exporter_type
    if resource_attributes is not None:
        os.environ["ANSYS_DT_OTEL__RESOURCE_ATTRIBUTES"] = str(resource_attributes)
    if headers is not None:
        os.environ["ANSYS_DT_OTEL__HEADERS"] = str(headers)



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for managing binary files.

This module also handles processes related to the Ansys HPS data transfer client.
"""

import json
import logging
import os
import platform
import stat
import subprocess
import threading
import time

import portend

from .exceptions import BinaryError
from .token import prepare_token

log = logging.getLogger(__name__)

level_map = {
    "trace": logging.DEBUG,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "error": logging.ERROR,
    "fatal": logging.CRITICAL,
    "panic": logging.CRITICAL,
}

verbosity_map = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
    3: logging.DEBUG,
}


def get_log_level(verbosity: int, debug: bool = False) -> int:
    """Get the log level based on verbosity and debug flag."""
    if debug:
        return logging.DEBUG
    return verbosity_map.get(verbosity, logging.INFO)


class PrepareSubprocess:
    """Provides for letting the context manager disable ``vfork`` and ``posix_spawn`` in the subprocess."""

    def __init__(self):
        """Initialize the PrepareSubprocess class object."""
        # Check if not Windows
        self.disable_vfork = os.name != "nt" and platform.system() != "Windows"
        self._orig_use_vfork = None
        self._orig_use_pspawn = None

    def __enter__(self):
        """Disable vfork and posix_spawn in subprocess."""
        if not self.disable_vfork:
            return

        if hasattr(subprocess, "_USE_VFORK"):
            self._orig_use_vfork = subprocess._USE_VFORK
            subprocess._USE_VFORK = False

        if hasattr(subprocess, "_USE_POSIX_SPAWN"):
            self._orig_use_pspawn = getattr(subprocess, "_USE_POSIX_SPAWN", False)
            subprocess._USE_POSIX_SPAWN = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Restore original values of _USE_VFORK and _USE_POSIX_SPAWN."""
        if not self.disable_vfork:
            return

        if self._orig_use_vfork is not None:
            subprocess._USE_VFORK = self._orig_use_vfork

        if self._orig_use_pspawn is not None:
            subprocess._USE_POSIX_SPAWN = self._orig_use_pspawn


def default_log_message(debug: bool, data: dict[str, any]):
    """Default log message handler.

    Parameters
    ----------
    data : dict
        Data to log.
    """
    # log.warning(f"Worker: {d}")

    level = data.pop("level", "info")
    data.pop("time", None)
    if not debug:
        data.pop("caller", None)
        data.pop("mode", None)
    msg = data.pop("message", None)

    if msg is None:
        return

    msg = msg.capitalize()
    level_no = level_map.get(level, logging.INFO)
    other = ""
    for k, v in data.items():
        formatted_value = f'"{v}"' if isinstance(v, str) and " " in v else v
        other += f"{k}={formatted_value} "
    other = other.strip()
    if other:
        msg += f" {other}"
    msg = msg.encode("ascii", errors="ignore").decode().strip()
    log.log(level_no, f"File Transfer: {msg}")


class BinaryConfig:
    """Provides for configuring the worker binary connection to the HPS data transfer client.

    Parameters
    ----------
    data_transfer_url: str, default: `https://localhost:8443/hps/dt/api/v1`
        Data transfer URL.
    log: bool, default: True
        Whether to enable logging.
    log_to_file: bool, default: False
        Whether to enable logging to a file.
    monitor_interval: float, default: 0.5
        Duration for waiting before the next monitor check on the binary.
    token: str
        Worker configuration setting of the access token credential.
    host: str, default: `127.0.0.1`
        Host IP to talk to the data transfer service.
    port: int
        Host port to talk to the data transfer service.
    verbosity: int, default: 1
        Verbosity level of the worker. The higher the number, the more verbose the logging.
    insecure: bool, default: False
        Whether to ignore SSL certificate verification.
    debug: bool, default: False
        Whether to enable debug logging.
    max_restarts: int, default: 5
        Maximum number of times to restart the worker if it crashes.
    """

    def __init__(
        self,
        # Required
        data_transfer_url: str = "https://localhost:8443/hps/dt/api/v1",
        # Process related settings
        # log: bool = True,
        log_to_file: bool = False,
        log_message: callable = default_log_message,
        monitor_interval: float = 0.5,
        # TODO: Remove path? not used anywhere
        path=None,
        # Worker config settings
        token: str = None,
        host: str = "127.0.0.1",
        port: int = None,
        verbosity: int = 3,
        insecure: bool = False,
        debug: bool = False,
        auth_type: str = None,
        env: dict | None = None,
        max_restarts: int = 5,
    ):
        """Initialize the BinaryConfig class object."""
        self.data_transfer_url = data_transfer_url

        # Process related settings
        self.log = debug
        self.log_to_file = log_to_file
        self._log_message = log_message
        self.monitor_interval = monitor_interval
        self.path = path

        # Worker config settings
        self.debug = debug
        self.data_transfer_url = data_transfer_url
        self.verbosity = verbosity
        self.host = host
        self._selected_port = port
        self._detected_port = None
        self._token = token
        self._env = env or {}
        self.insecure = insecure
        self.auth_type = auth_type
        self.max_restarts = max_restarts

        self._on_token_update = None
        self._on_process_died = None
        self._on_port_changed = None

    def update(self, **kwargs):
        """Update worker configuration settings."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError(f"Unknown attribute {key}")
        if self.debug:
            self.log = True

    @property
    def port(self):
        """Port."""
        return self._selected_port or self._detected_port

    @port.setter
    def port(self, value):
        """Set port."""
        self._selected_port = value

    @property
    def token(self):
        """Token."""
        return self._token

    @token.setter
    def token(self, value):
        """Set token."""
        if self.debug:
            log.debug(
                f"Setting token to ...{value[-10:]}, old token: {f'...{self._token[-10:]}' if self._token else 'none'}"
            )
        self._token = value
        if self._on_token_update is not None:
            self._on_token_update()

    @property
    def url(self):
        """URL."""
        return f"http://{self.host}:{self.port}/api/v1"

    @property
    def env(self):
        """Get additional environment variables that will be passed to the child process."""
        return self._env

    @env.setter
    def env(self, value):
        """Set environment variables to pass to the child process."""
        if not isinstance(value, dict):
            raise TypeError("Environment variables must be a dictionary.")
        self._env = value

    @property
    def log_message(self):
        """Get the log message handler."""
        return self._log_message

    @log_message.setter
    def log_message(self, value):
        """Set the log message handler."""
        if not callable(value):
            raise TypeError("Log message handler must be a callable.")
        self._log_message = value


class Binary:
    """Provides for starting, stopping, and monitoring the worker binary.

    Parameters
    ----------
    config: BinaryConfig
        Binary configuration.
    """

    def __init__(
        self,
        config: BinaryConfig | None = None,
    ):
        """Initialize the Binary class object."""
        self._config = config or BinaryConfig()

        self._base_args = []
        self._args = []
        self._stop = None
        self._prepared = None
        self._process = None
        self._log_thread = None

    def __getstate__(self):
        """Return state of the object."""
        state = self.__dict__.copy()
        del state["_stop"]
        del state["_prepared"]
        del state["_process"]
        del state["_log_thread"]
        return state

    @property
    def config(self):
        """Configuration."""
        return self._config

    @property
    def is_started(self):
        """Flag indicating if the binary is up and running."""
        try:
            return self._process is not None and self._process.returncode is None
        except Exception:
            return False

    def start(self):
        """Start the worker binary.

        This method checks for the binary in a set path, marks the binary as an executable,
        and then starts the executable.
        """
        if self._process is not None and self._process.returncode is None:
            raise BinaryError("Worker already started.")

        log.debug("Starting worker ...")

        self._stop = threading.Event()
        self._prepared = threading.Event()

        bin_path = self._config.path
        if not bin_path or not os.path.exists(bin_path):
            raise BinaryError(f"Binary not found: {bin_path}")

        # Mark binary as executable
        if not os.access(bin_path, os.X_OK):
            # log.debug(f"Marking binary as executable: {bin_path}")
            st = os.stat(bin_path)
            os.chmod(bin_path, st.st_mode | stat.S_IEXEC)

        self._process = None

        # if False:
        t = threading.Thread(target=self._monitor, args=(), name="worker_monitor")
        t.daemon = True
        t.start()

        if not self._prepared.wait(timeout=5.0):
            log.warning("Worker preparation is taking longer than expected ...")

    def stop(self, wait=5.0):
        """Stop the worker binary."""
        if self._process is None:
            return

        self._stop.set()
        self._prepared.clear()

        start = time.time()
        while True:
            if self._process.poll() is not None:
                log.debug("Worker stopped.")
                break
            if time.time() - start > wait:
                log.warning("Worker did not stop in time, killing ...")
                self._process.kill()
                break
            time.sleep(wait * 0.1)

    def _log_output(self):
        log_message = self._config._log_message
        # Keep track of the first time the process is present,
        # if it disappears, we stop reading, it means the process has ended
        started = False
        while not self._stop.is_set():
            if self._process is None or self._process.stdout is None:
                if started:
                    log.debug("Log thread found the process stdout missing, reading stopped.")
                    break
                time.sleep(1)
                continue
            try:
                started = True
                line = self._process.stdout.readline()
                if not line:
                    log.debug("Log thread stdout ended normally, reading stopped.")
                    break
                # If we dont need to log it, it still needs to be read to prevent hangs from full buffers
                if not self.config.log:
                    continue
                line = line.decode(errors="strip").strip()
                if log_message is not None:
                    d = json.loads(line)
                    log_message(self._config.debug, d)
            except json.decoder.JSONDecodeError:
                pass
            except Exception as e:
                if self._config.debug:
                    log.debug(f"Error reading worker output: {e}")
                time.sleep(1)
        # log.debug("Worker log output stopped")

    def _monitor(self):
        restart_count = 0  # Initialize a counter for restarts
        while not self._stop.is_set():
            if self._process is None:
                log.info(f"Data Transfer is starting on restart counter {restart_count}")
                self._prepare()
                args = " ".join(self._args)

                redacted = f"{args}"
                if self._config.token is not None:
                    redacted = args.replace(self._config.token, "***")

                env = os.environ.copy()
                env_str = ""
                if self._config.env:
                    env.update(self._config.env)
                    env_str = ",".join([k for k in self._config.env.keys() if k != "PATH"])

                log.debug(f"Command: {redacted}")
                if self._config.debug:
                    log.debug(f"Environment: {env_str}")

                with PrepareSubprocess():
                    log.info("Launching data transfer worker")
                    self._process = subprocess.Popen(
                        args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env
                    )
                    log.info(f"Data transfer worker is running with PID: {self._process.pid}")

                    self._log_thread = threading.Thread(target=self._log_output, args=(), name="worker_log_output")
                    self._log_thread.daemon = True
                    self._log_thread.start()
            else:
                ret_code = self._process.poll()
                if ret_code is not None and ret_code != 0:
                    restart_count += 1  # Increment the restart counter
                    if restart_count > self.config.max_restarts:
                        log.error(f"Worker exceeded maximum restart attempts ({self.config.max_restarts}). Stopping...")
                        break  # Exit the loop after exceeding the restart limit

                    log.warning(f"Worker exited with code {ret_code}, restarting ...")
                    self._process = None
                    self._prepared.clear()
                    if self.config._on_process_died is not None:
                        self.config._on_process_died(ret_code)

                    if self._log_thread is not None:
                        log.info("Waiting for log thread to finish ...")
                        self._log_thread.join()
                        self._log_thread = None
                    time.sleep(1.0)
                    continue
            # Reset restart_count if the worker is running successfully
            restart_count = 0

            time.sleep(self._config.monitor_interval)
        log.debug("Worker monitor stopped")

    def _prepare(self):
        if self._config._selected_port is None:
            self._config._detected_port = self._get_open_port()
            if self._config._on_port_changed is not None:
                self._config._on_port_changed(self._config._detected_port)

        self._build_base_args()

        self._build_args()
        self._prepared.set()

    def _get_open_port(self):
        port = portend.find_available_local_port()
        return port

    def _build_base_args(self):
        log_types = ["diode"]
        if self._config.log_to_file:
            log_types.append("file")

        self._base_args = [
            self._config.path,
            "--log-types",
            ",".join(log_types),
            "--host",
            self._config.host,
            "--port",
            str(self._config.port),
        ]

    def _build_args(self):
        self._args = list(self._base_args)
        self._args.extend(
            [
                "--dt-url",
                self._config.data_transfer_url,
            ]
        )

        # self._args.extend(
        #     [
        #         "-v",
        #         str(self._config.verbosity),
        #     ]
        # )

        if self._config.insecure:
            self._args.append("--insecure")

        if self._config.debug:
            self._args.append("--debug")

        if self._config.auth_type:
            self._args.extend(
                [
                    "--auth-type",
                    self._config.auth_type,
                ]
            )

        if self._config.token is not None:
            self._args.extend(
                [
                    "-t",
                    f'"{prepare_token(self._config.token)}"',
                ]
            )



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the base class for all client and server HPS-related errors."""

import httpx
from requests.exceptions import RequestException


class HPSError(RequestException):
    """Provides the base class for all HPS-related errors.

    This class derives from the :class:`requests.exceptions.RequestException`
    base class.

    Example:
        >>> from ansys.hps.client import HPSError
        >>> from ansys.hps.client.jms import Client
        >>> try:
        >>>     client = Client(url="https://127.0.0.1:8443/hps/",
                                username="repuser",
                                password="wrong_psw")
        >>> except HPSError as e:
        >>>     print(e)
        401 Client Error: invalid_grant for: POST https://127.0.0.1:8443/hps/auth...
        Invalid user credentials
    """

    def __init__(self, *args, **kwargs):
        """Initializes the HPSError class."""
        self.reason = kwargs.pop("reason", None)
        self.description = kwargs.pop("description", None)
        self.give_up = kwargs.pop("give_up", False)
        super().__init__(*args, **kwargs)


class APIError(HPSError):
    """Provides server-side related errors."""

    def __init__(self, *args, **kwargs):
        """Initializes the APIError class object."""
        super().__init__(*args, **kwargs)


class ClientError(HPSError):
    """Provides client-side related errors."""

    def __init__(self, *args, **kwargs):
        """Initializes the ClientError class object."""
        super().__init__(*args, **kwargs)


class BinaryError(HPSError):
    """Provides binary-related errors."""

    def __init__(self, *args, **kwargs):
        """Initializes the BinaryError class object."""
        super().__init__(*args, **kwargs)


class NotReadyError(ClientError):
    """Provides not ready-related errors."""

    def __init__(self, *args, **kwargs):
        """Initializes the NotReadyError class object."""
        super().__init__(*args, **kwargs)


class TimeoutError(ClientError):
    """Provides timeout-related errors."""

    def __init__(self, *args, **kwargs):
        """Initializes the TimeoutError class object."""
        super().__init__(*args, **kwargs)


def _raise_for_status(response: httpx.Response):
    r_content = {}
    try:
        r_content = response.json()
    except ValueError:
        pass

    reason = r_content.get("title", None)  # jms api
    if not reason:
        reason = r_content.get("error", None)  # auth api
    if not reason:
        reason = getattr(response, "reason", None)

    description = r_content.get("description", None)  # jms api
    if not description:
        description = r_content.get("error_description", None)  # auth api

    if response.status_code == 425:
        raise NotReadyError(
            response.status_code,
            reason=reason,
            description=description,
            response=response,
        )

    if 400 <= response.status_code < 500:
        error_msg = f"{response.status_code} Client Error: {reason} for: {response.request.method} {response.url}"
        if description:
            error_msg += f"\n{description}"

        give_up = response.status_code in [401, 403]
        raise ClientError(error_msg, reason=reason, description=description, response=response, give_up=give_up)
    elif 500 <= response.status_code < 600:
        error_msg = f"{response.status_code} Server Error: {reason} for: {response.request.method} {response.url}"
        if description:
            error_msg += f"\n{description}"
        raise APIError(error_msg, reason=reason, description=description, response=response)
    return response


def raise_for_status(response: httpx.Response):
    """Automatically check for HTTP errors.

    This method mimics the ``requests.Response.raise_for_status()`` method.
    """
    if response.status_code < 400 or response.status_code >= 600:
        return

    if getattr(response, "is_error", False):
        response.read()

    _raise_for_status(response)


async def async_raise_for_status(response: httpx.Response):
    """Automatically check for HTTP errors.

    This method mimics the ``requests.Response.raise_for_status()`` method.
    """
    if response.status_code < 400 or response.status_code >= 600:
        return

    if getattr(response, "is_error", False):
        await response.aread()

    _raise_for_status(response)



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides utilities for implementing exponential backoff with optional jitter.

This module is commonly used in retry mechanisms to handle transient errors in distributed systems.
"""

from random import uniform


def get_expo_backoff(
    base: float, attempts: int = 1, cap: float = 100_000_000, attempts_cap: int = 100_000_000, jitter: bool = True
):
    """Get the backoff value.

    For more information, see `Exponential Backoff And Jitter
    <https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/>`_
    in the AWS documentation.

    Parameters
    ==========
    base : float
        Time to sleep in the first attempt.
    attempts : int, default: 1
        Number of attempts that have already been made.
    cap : float, default: 100000000
        Maximum value that can be returned.
    jitter : bool, default: True
        Whether to apply jitter to the returned value.
    """
    # Full jitter formula
    # https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    attempts = min(attempts, attempts_cap)
    if jitter:
        try:
            return uniform(base, min(cap, base * 2 ** (attempts - 1)))
        except OverflowError:
            return cap
    else:
        return min(cap, base * 2 ** (attempts - 1))



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the Python client to the HPS data transfer APIs."""

import asyncio
import atexit
from collections.abc import Awaitable, Callable
import logging
import os
import platform
import random
import shutil
import stat
import string
import threading
import time
import traceback

import backoff
import filelock
import httpx
import humanize as hz
import psutil
import urllib3

from ansys.hps.data_transfer.client.binary import Binary, BinaryConfig
from ansys.hps.data_transfer.client.exceptions import BinaryError, async_raise_for_status, raise_for_status
from ansys.hps.data_transfer.client.token import prepare_token

urllib3.disable_warnings()

for n in ["httpx", "httpcore", "requests", "urllib3", "filelock"]:
    logger = logging.getLogger(n)
    logger.setLevel(logging.CRITICAL)

api_key_header_env = "ANSYS_DT_AUTH__API_KEY__HEADER_NAME"
api_key_value_env = "ANSYS_DT_AUTH__API_KEY__VALUE"

log = logging.getLogger(__name__)


def bin_in_use(bin_path):
    """Check if a binary is in use."""
    for proc in psutil.process_iter():
        try:
            cmd = proc.cmdline()
            for c in cmd:
                if bin_path in c:
                    return True
        except psutil.NoSuchProcess:
            pass
        except psutil.AccessDenied:
            pass
        except Exception as err:
            log.debug(f"Error checking process: {err}")
    return False


def flatten_features(y, separator="."):
    """Flatten a nested dictionary into a list of strings."""
    out = []

    def flatten(x, name=""):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + separator)
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + separator)
                i += 1
        else:
            s = name[:-1] + str(x)
            s = s.replace(" ", "_").replace("-", "_")
            out.append(s)

    flatten(y)
    return out


class MonitorState:
    """Provides for monitoring and tracking the state of the worker binary."""

    def __init__(self):
        """Initialize the MonitorState class object."""
        self.reset()
        self._sleep_not_started = 2
        self._sleep_while_running = 5

    @property
    def sleep_for(self):
        """Sleep time based on the worker state."""
        return self._sleep_while_running if self._was_ready else self._sleep_not_started

    def reset(self):
        """Reset the monitor state to the initial values."""
        self._ok_reported = False
        self._was_ready = False
        self._failed = False
        self._last_exc = None

    def mark_ready(self, ready):
        """Mark the worker as ready or not ready."""
        self._was_ready = True
        msg = f"Worker is running, reporting {'' if ready else 'not '}ready"
        if ready:
            if not self._ok_reported:
                log.debug(msg)
                self._ok_reported = True
        else:
            self._ok_reported = False
            log.debug(msg)

    def mark_failed(self, exc=None, binary=None):
        """Mark the worker as failed."""
        exc_str = "" if exc is None else f": {exc}"
        if binary:
            bin_str = f", binary is {'' if binary.is_started else 'not '}running"
        else:
            bin_str = ""
        log.warning(f"Worker failure detected{exc_str}{bin_str}")

        self._ok_reported = False
        self._failed = True
        self._last_exc = exc

    def report(self, binary):
        """Report the worker status."""
        if self._failed and self._was_ready:
            descr = "running" if binary.is_started else "not running"
            if self._last_exc is not None:
                descr += f", last exception: {self._last_exc}"
            log.warning(f"Worker failure detected, binary is {descr}")


class ClientBase:
    """Provides the Python client to the HPS data transfer APIs.

    This class uses the provided credentials to create and store
    an authorized :class:`requests.Session` object.

    Parameters
    ----------
    bin_config: BinaryConfig, default: None
        Binary configuration. If no configuration is provided, a default ``BinaryConfig`` object is created.
    download_dir: str, default: "dt_download"
        Path to the download directory.
    clean: bool, default: False
        Whether to clean the path to the download directory.
    clean_dev: bool, default: True
        Whether to clean the path to the download directory if the binary is from the development branch.
    check_in_use: bool, default: True
        Whether to check if the binary is in use and skip downloading a new binary.
    refresh_token_callback: Callable[[], str], default: None
        Callback function to refresh the access token. This function should return a new access token.
    timeout: float, default: 60.0
        Timeout for the session. This is the maximum time to wait for a response from the server.
    retries: int, default: 1
        Number of times to retry the operation.

    Examples:
    --------
    Create a client object and connect to HPS data transfer with an access token.

    >>> from ansys.hps.data_transfer.client import Client
    >>> def refresh_token_cb():
            # Function to refresh the token
            token = authenticate(username=username, password, verify=False, url=auth_url)
            token = token.get("access_token", None)
            return token
    >>> client = Client(clean=True)
    >>> client.refresh_token_callback = refresh_token_cb
    >>> client.binary_config.update(
            verbosity=3,
            debug=False,
            insecure=True,
            token=token,
            data_transfer_url=dt_url,
        )
    >>> client.start()

    """

    class Meta:
        """Meta class for the ``ClientBase`` class."""

        is_async = False

    def __init__(
        self,
        bin_config: BinaryConfig = None,
        download_dir: str = "dt_download",
        clean=False,
        clean_dev=True,
        check_in_use=True,
        refresh_token_callback: Callable[[], str | Awaitable[str]] = None,
        timeout=5.0,
        retries=4,
    ):
        """Initializes the Client class object."""
        self._bin_config = bin_config or BinaryConfig()
        self._download_dir = download_dir
        self._clean = clean
        self._clean_dev = clean_dev
        self._check_in_use = check_in_use
        self._timeout = timeout
        self.retries = retries
        self.refresh_token_callback = refresh_token_callback

        self._session = None
        self.binary = None
        self.panic_file = None

        self._features = None
        self._api_key = None
        self._api_key_header = "X-API-Key"

        self._monitor_stop = None
        self._monitor_state = MonitorState()

        self.progress_interval = 5.0  # seconds
        self._unauthorized_max_retry = 1
        self._unauthorized_num_retry = 0

    def __getstate__(self):
        """Return pickled state of the object."""
        state = self.__dict__.copy()
        del state["_session"]
        del state["_monitor_stop"]
        return state

    def __setstate__(self, state):
        """Restore state from pickled state."""
        self.__dict__.update(state)
        self._session = None
        self._monitor_stop = None

    @property
    def unauthorized_max_retry(self):
        """Getter for unauthorized_max_retry."""
        return self._unauthorized_max_retry

    @unauthorized_max_retry.setter
    def unauthorized_max_retry(self, value):
        """Setter for unauthorized_max_retry."""
        self._unauthorized_max_retry = value

    @property
    def binary_config(self):
        """Binary configuration."""
        return self._bin_config

    @property
    def pid(self):
        """Process ID of the worker binary."""
        if self.binary is None or self.binary._process is None:
            return None
        return self.binary._process.pid

    @property
    def base_api_url(self):
        """API URL from the configuration."""
        return self._bin_config.url

    @property
    def session(self):
        """Session object. If one does not exist, a new one is created."""
        if self._session is None:
            self._session = self._create_session(self.base_api_url, sync=not self.Meta.is_async)
        return self._session

    @property
    def is_started(self):
        """Flag indicating if the binary is up and running."""
        return self.binary is not None and self.binary.is_started

    @property
    def timeout(self):
        """Timeout for the session."""
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        """Set the timeout for the session."""
        self._timeout = value

    @property
    def retries(self):
        """Number of retries."""
        return self._retries

    @retries.setter
    def retries(self, value):
        """Set the number of retries."""
        self._retries = value

    def start(self):
        """Start the client session using the binary configuration credentials."""
        if self.binary is not None:
            return

        if self._clean and os.path.exists(self._download_dir):
            try:
                shutil.rmtree(self._download_dir)
            except Exception as ex:
                log.debug(f"Failed to remove directory {self._download_dir}: {ex}")

        self._monitor_stop = threading.Event()
        self._monitor_state.reset()
        self._prepare_platform_binary()

        self._bin_config._on_port_changed = self._on_port_changed
        self._bin_config._on_process_died = self._on_process_died

        self._adjust_config()
        self.binary = Binary(config=self._bin_config)
        self.binary.start()

        # self._session = self._create_session(self.base_api_url)

    def stop(self, wait=5.0):
        """Stop the client session."""
        if self.binary is None:
            return
        self._monitor_stop.set()
        self.binary.stop(wait=wait)
        self.binary = None
        self._session = None
        self._bin_config._on_port_changed = None

    def has(self, feature):
        """Check if the feature is available using a dot notation."""
        if self._features is None:
            return False
        return feature in self._features

    def _platform(self):
        plat = ""
        match platform.system():
            case "Windows":
                plat = "win"
            case "Linux":
                plat = "lin"
            case "Darwin":
                plat = "darwin"

        if not plat:
            raise BinaryError(f"Unsupported platform: {platform.system()}")

        arch = ""
        if plat == "win":
            match platform.uname().machine:
                case "AMD64":
                    arch = "x64"
        else:
            match os.uname().machine:
                case "x86_64":
                    arch = "x64"
                case "aarch64":
                    arch = "arm64"
                case "arm64":
                    arch = "arm64"

        if not arch:
            raise BinaryError(f"Unsupported architecture: {os.uname().machine}")

        return f"{plat}-{arch}"

    def _prepare_bin_path(self, build_info):
        log.debug("Server build info:")
        for k, v in build_info.items():
            log.debug(f"  {k}: {v}")
        version_hash = build_info["version_hash"]

        # Figure out binary download path
        bin_ext = ".exe" if platform.system() == "Windows" else ""
        bin_dir = os.path.join(self._download_dir, "worker")
        bin_path = os.path.join(bin_dir, f"hpsdata-{version_hash}{bin_ext}")
        return bin_path

    def _get_features(self, d):
        f = d.get("features", None)
        if f is None:
            self._features = None
            return
        self._features = flatten_features(f)

    def _check_binary(self, build_info, bin_path):
        """Check if there is a need to download the binary."""
        branch = build_info["branch"]

        # Check if we need to download the binary
        reason = None
        if self._bin_config.path is not None and not os.path.exists(self._bin_config.path):
            reason = "binary not found"
        elif not os.path.exists(bin_path):
            reason = "binary version not found"
        elif self._clean_dev and branch == "dev":
            reason = "dev branch"

        # Use downloaded binary if nothing else was specified
        if self._bin_config.path is None and os.path.exists(bin_path):
            self._bin_config.path = bin_path

        return reason, bin_path

    def _prepare_platform_binary(self):
        # Get service build info
        dt_url = self._bin_config.data_transfer_url
        session = self._create_session(dt_url, sync=True)
        resp = session.get("/")
        if resp.status_code != 200:
            raise BinaryError(f"Failed to download binary: {resp.text}")

        d = resp.json()

        self._get_features(d)
        bin_path = self._prepare_bin_path(d["build_info"])
        lock_name = f"{os.path.splitext(os.path.basename(bin_path))[0]}.lock"
        lock_dir = os.path.dirname(bin_path)
        if not os.path.exists(lock_dir):
            try:
                os.makedirs(lock_dir)
            except Exception as ex:
                log.debug(f"Failed to create lock dir: {ex}")
        lock_path = os.path.join(lock_dir, lock_name)
        lock = filelock.SoftFileLock(lock_path, timeout=60)

        try:
            with lock:
                reason, bin_path = self._check_binary(d["build_info"], bin_path)
                bin_path = os.path.abspath(bin_path)

                if self._check_in_use and bin_in_use(bin_path):
                    log.info(f"Skipping download, binary in use: {bin_path}")
                    return

                if reason is None:
                    log.debug(f"Using existing binary: {self._bin_config.path}")
                    return

                bin_dir = os.path.dirname(bin_path)
                bin_ext = os.path.splitext(bin_path)[1]
                if not os.path.exists(bin_dir):
                    try:
                        os.makedirs(bin_dir)
                    except Exception as ex:
                        log.warning(f"Failed to create directory {bin_dir}: {ex}")

                platform_str = self._platform()
                log.info(
                    f"Downloading data transfer worker for platform '{platform_str}' from {dt_url}, reason: {reason}"
                )
                log.debug(f"Binary download path: {bin_path}")
                url = f"/binaries/worker/{platform_str}/hpsdata{bin_ext}"
                try:
                    start = time.time()
                    last_progress = start
                    written = 0
                    with open(bin_path, "wb") as f, session.stream("GET", url) as resp:
                        resp.read()
                        if resp.status_code != 200:
                            raise BinaryError(f"Failed to download binary: {resp.text}")

                        for chunk in resp.iter_bytes():
                            now = time.time()
                            written += f.write(chunk)

                            since_last_progress = now - last_progress
                            if since_last_progress > self.progress_interval:
                                msg = f"Downloading binary, {hz.precisedelta(now - start)} so far ..."
                                content_length = resp.headers.get("Content-Length", None)
                                if content_length is not None:
                                    content_length = int(content_length)
                                    prog = float(written) / float(content_length) * 100.0
                                    msg += f" {prog:.1f}%, {hz.naturalsize(written)}/{hz.naturalsize(content_length)}"
                                else:
                                    msg += f" {hz.naturalsize(written)} downloaded"
                                log.info(msg)
                                last_progress = now

                    self._bin_config.path = bin_path
                except Exception as ex:
                    if self._bin_config.debug:
                        log.debug(traceback.format_exc())
                    log.error(f"Failed to download binary: {ex}")
                    os.remove(bin_path)

                st = os.stat(bin_path)
                # log.debug(f"Marking binary as executable: {bin_path}")
                os.chmod(bin_path, st.st_mode | stat.S_IEXEC)
                # if self._bin_config.debug:
                # log.debug(f"Binary mode: {stat.filemode(os.stat(bin_path).st_mode)}")
        except filelock.Timeout as ex:
            raise BinaryError(f"Failed to acquire lock for binary download: {lock_path}") from ex

    def _create_session(self, url: str, *, sync: bool = True):
        verify = not self._bin_config.insecure
        # log.debug("Creating session for %s with verify=%s", url, verify)

        args = {
            "timeout": httpx.Timeout(self._timeout),
        }

        if sync:
            session = httpx.Client(
                transport=httpx.HTTPTransport(retries=self._retries, verify=verify),
                event_hooks={"response": [self._auto_refresh_token, raise_for_status]},
                **args,
            )
            session._transport.verify = False
        else:
            session = httpx.AsyncClient(
                transport=httpx.AsyncHTTPTransport(retries=self._retries, verify=verify),
                event_hooks={"response": [self._async_auto_refresh_token, async_raise_for_status]},
                **args,
            )
        session.base_url = url
        session.verify = verify
        session.follow_redirects = True

        if self._bin_config.token is not None:
            session.headers["Authorization"] = prepare_token(self._bin_config.token)
        if self._api_key:
            session.headers[self._api_key_header] = self._api_key

        return session

    def _on_port_changed(self, port):
        log.debug(f"Port changed to {port}")
        if self._session is not None:
            self._session.close()
        self._session = None

    def _on_process_died(self, ret_code):
        self._monitor_state.reset()

    def _auto_refresh_token(self, response: httpx.Response):
        """Provide a callback for refreshing an expired token.

        Automatically refreshes the access token and
        re-sends the request in case of an unauthorized error.
        """
        log.debug(f"response status: {response.status_code} for {response.request.method} {response.url}")
        if response.status_code == 401 and self._unauthorized_num_retry < self._unauthorized_max_retry:
            log.info("401 authorization error: Trying to get a new access token.")
            if self.refresh_token_callback is None:
                log.info("No refresh callback provided, skipping token refresh.")
                return
            self._unauthorized_num_retry += 1
            token = self.refresh_token_callback()
            self._bin_config.token = token
            # Update the Authorization header
            response.request.headers.update({"Authorization": f"Bearer {token}"})
            log.debug("Retrying request with updated access token.")
            retried_response = self._session.send(response.request)
            log.debug(f"Retried response status: {retried_response.status_code}")
            # Modify the response body
            response._content = retried_response.content
            response.status_code = retried_response.status_code
            return

        self._unauthorized_num_retry = 0
        return

    async def _async_auto_refresh_token(self, response: httpx.Response):
        """Provide a callback for refreshing an expired token.

        Automatically refreshes the access token and
        re-sends the request in case of an unauthorized error.
        """
        log.debug(f"response status: {response.status_code} for {response.request.method} {response.url}")
        if response.status_code == 401 and self._unauthorized_num_retry < self._unauthorized_max_retry:
            log.info("401 authorization error: Trying to get a new access token.")
            if self.refresh_token_callback is None:
                log.info("No refresh callback provided, skipping token refresh.")
                return
            self._unauthorized_num_retry += 1
            token = await self.refresh_token_callback()
            self._bin_config.token = token
            # Update the Authorization header
            response.request.headers.update({"Authorization": f"Bearer {token}"})
            log.debug("Retrying request with updated access token.")
            retried_response = await self._session.send(response.request)
            log.debug(f"Retried response status: {retried_response.status_code}")
            # Modify the response body
            response._content = retried_response.content
            response.status_code = retried_response.status_code
            return

        self._unauthorized_num_retry = 0
        return

    def _adjust_config(self):
        if not self._features:
            return

        if self.has("auth_types.api_key"):
            self._bin_config.auth_type = "api-key"
            self._api_key = "".join(
                random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(128)
            )
            env = {
                api_key_header_env: self._api_key_header,
                api_key_value_env: self._api_key,
            }
            self._bin_config.env.update({k: v for k, v in env.items() if k not in os.environ})

    def _fetch_panic_file(self, resp):
        """Extract and log the panic file location from the response."""
        if resp.status_code == 200:
            self.panic_file = resp.json().get("debug", {}).get("panic_file", None)
            log.debug(f"Worker panic file: {self.panic_file}")

    def _panic_file_contents(self):
        """Read and log the contents of the panic file if it exists."""
        # if the file exists and the size of the file is > 0,
        # read and log its content
        if self.panic_file and os.path.exists(self.panic_file):
            try:
                if os.path.getsize(self.panic_file) > 0:
                    with open(self.panic_file) as f:
                        # Read the file line by line
                        lines = f.readlines()
                        message = []
                        for line in lines:
                            # Check for empty lines to split the message
                            if line.strip() == "":
                                log.error(f"Worker panic file content:\n{''.join(message)}")
                                message = []  # Reset the message buffer
                            else:
                                message.append(line)

                        # Log any remaining content after the last empty line
                        if message:
                            log.error(f"Worker panic file content:\n{''.join(message)}")
            except Exception as panic_ex:
                log.debug(f"Failed to read panic file: {panic_ex}")


class AsyncClient(ClientBase):
    """Provides an async interface to the Python client to the HPS data transfer APIs."""

    class Meta(ClientBase.Meta):
        """Meta class for AsyncClient class."""

        is_async = True

    def __init__(self, *args, **kwargs):
        """Initializes the AsyncClient class object."""
        super().__init__(*args, **kwargs)
        self._bin_config._on_token_update = self._update_token
        self._monitor_task = None

    def __getstate__(self):
        """Return pickled state of the object."""
        state = super().__getstate__()
        del state["_monitor_task"]
        return state

    def __setstate__(self, state):
        """Restore state from pickled state."""
        super().__setstate__(state)
        self.__dict__.update(state)
        self._monitor_task = None

    async def start(self):
        """Start the async binary worker."""
        super().start()
        # grab location of panic file
        resp = await self.session.get("/")
        self._fetch_panic_file(resp)
        self._monitor_task = asyncio.create_task(self._monitor())

    async def stop(self, wait=5.0):
        """Stop the async binary worker."""
        if self._session is not None:
            try:
                await self._session.post(self.base_api_url + "/shutdown")
                await asyncio.sleep(0.1)
                if self._monitor_task is not None:
                    self._monitor_task.cancel()
            except Exception as ex:
                log.warning(f"Failed to send shutdown request: {ex}")
            finally:
                await self._session.aclose()  # Ensure the session is closed
                self._session = None
        super().stop(wait=wait)
        # asyncio_atexit.register(self.stop)

    async def wait(self, timeout: float = 60.0, sleep=0.5):
        """Wait on the async binary worker."""
        start = time.time()
        while time.time() - start < timeout:
            try:
                if self._session is not None:
                    resp = await self.session.get(self.base_api_url)
                    if resp.status_code != 200:
                        log.debug("Waiting for worker to start")
                    else:
                        return
            except Exception as ex:
                if self._bin_config.debug:
                    log.debug(f"Error waiting for worker to start: {ex}")
            finally:
                await asyncio.sleep(backoff.full_jitter(sleep))

    def _update_token(self):
        loop = asyncio.get_running_loop()
        if self._session is None:
            return
        log.debug("Updating auth token, ends in %s", self._bin_config.token[-10:])
        try:
            self._session.headers["Authorization"] = prepare_token(self._bin_config.token)
            # Make sure the token gets intercepted by the worker
            loop.run_until_complete(self.session.get("/"))
        except Exception as e:
            log.debug(f"Error updating token: {e}")

    async def _monitor(self):
        while True:
            try:
                await asyncio.sleep(self._monitor_state.sleep_for)
                if self._session is None or self.binary is None:
                    continue

                resp = await self._session.get(self.base_api_url)

                if resp.status_code == 200:
                    ready = resp.json().get("ready", False)
                    self._monitor_state.mark_ready(ready)
                    continue
            except asyncio.CancelledError:
                log.debug("Monitor task cancelled")
                return
            except Exception as ex:
                if self.binary_config.debug:
                    log.debug("URL: %s", self.base_api_url)
                    log.debug(traceback.format_exc())
                # Before marking it as failed check if there is a panic file
                self._panic_file_contents()
                self._monitor_state.mark_failed(exc=ex, binary=self.binary)
                continue

            self._monitor_state.report(self.binary)


class Client(ClientBase):
    """Provides the Python client to the HPS data transfer APIs.

    This class uses the provided credentials to create and store
    an authorized :class:`requests.Session` object.
    """

    class Meta(ClientBase.Meta):
        """Meta class for Client class."""

        is_async = False

    def __init__(self, *args, **kwargs):
        """Initializes the Client class object."""
        super().__init__(*args, **kwargs)
        self._bin_config._on_token_update = self._update_token
        self._monitor_thread = None

    def __getstate__(self):
        """Return pickled state of the object."""
        state = super().__getstate__()
        del state["_monitor_thread"]
        return state

    def __setstate__(self, state):
        """Restore state from pickled state."""
        super().__setstate__(state)
        self.__dict__.update(state)
        self._monitor_thread = None

    def start(self):
        """Start the client session using the binary configuration credentials."""
        super().start()
        atexit.register(self.stop)
        self._monitor_thread = threading.Thread(
            target=self._monitor, args=(), daemon=True, name="worker_status_monitor"
        )
        # grab location of panic file
        resp = self.session.get("/")
        self._fetch_panic_file(resp)
        self._monitor_thread.start()

    def stop(self, wait=5.0):
        """Stop the client session."""
        if self._session is not None:
            try:
                self._session.post(self.base_api_url + "/shutdown")
            except Exception as ex:
                log.warning(f"Failed to send shutdown request: {ex}")
        super().stop(wait=wait)

    def wait(self, timeout: float = 60.0, sleep=0.5):
        """Wait on the worker binary to start."""
        start = time.time()
        while time.time() - start < timeout:
            try:
                if self._session is not None:
                    resp = self._session.get(self.base_api_url)
                    if resp.status_code != 200:
                        log.debug("Waiting for worker to start")
                    else:
                        return
            except Exception as ex:
                if self._bin_config.debug:
                    log.debug(f"Error waiting for worker to start: {ex}")
            finally:
                time.sleep(backoff.full_jitter(sleep))

    def _update_token(self):
        if self._session is None:
            return
        log.debug("Updating auth token, ends in %s", self._bin_config.token[-10:])
        try:
            self._session.headers["Authorization"] = prepare_token(self._bin_config.token)
            # Make sure the token gets intercepted by the worker
            resp = self.session.get("/")
            if resp.status_code != 200:
                log.error(f"Failed to update token, server responded with: {resp.status_code}")
        except Exception as e:
            log.debug(f"Error updating token: {e}")

    def _monitor(self):
        while not self._monitor_stop.is_set():
            time.sleep(self._monitor_state.sleep_for)

            if self._session is None or self.binary is None:
                continue
            try:
                resp = self._session.get(self.base_api_url)

                if resp.status_code == 200:
                    ready = resp.json().get("ready", False)
                    self._monitor_state.mark_ready(ready)
                    continue
            except Exception as ex:
                if self.binary_config.debug:
                    log.debug("URL: %s", self.base_api_url)
                    log.debug(traceback.format_exc())
                # Before marking it as failed check if there is a panic file
                self._panic_file_contents()
                self._monitor_state.mark_failed(exc=ex, binary=self.binary)
                continue

            self._monitor_state.report(self.binary)
        log.debug("Worker status monitor stopped")



# Copyright (C) 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# generated by datamodel-codegen

from __future__ import annotations

from enum import Enum
from typing import Any

from pydantic import AnyUrl, AwareDatetime, BaseModel, ConfigDict, Field, conint


class AuthRedirectRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )


class BinaryInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    name: str | None = None
    platform: str | None = None
    size: int | None = None
    type: str | None = None


class BuildInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    architecture: str | None = None
    branch: str | None = None
    go_version: str | None = None
    mode: str | None = None
    operating_system: str | None = None
    revision: str | None = None
    short_revision: str | None = None
    timestamp: str | None = None
    version: str | None = None
    version_hash: str | None = None


class CheckPermissionsResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    allowed: bool | None = Field(default=None, description="Indicates if the check was successful")


class CompatBackend(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    name: str | None = None
    obj_type: str | None = None
    priority: int | None = None
    reference: str | None = None
    url: str | None = None
    use_default_url: bool | None = None


class CompatBuildInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    branch: str | None = None
    revision: str | None = None
    short_revision: str | None = None
    timestamp: str | None = None
    version: str | None = None


class CompatCopyRequestBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    destination: str | None = None


class CompatCopyResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    checksum: str | None = None


class CompatHealthResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    ansft: bool | None = None
    build: CompatBuildInfo | None = None


class CompatStorageResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    backends: list[CompatBackend] | None = None


class CompatUploadResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    checksum: str | None = None


class DataAssignment(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    compressed_size: conint(ge=0) | None = None
    compression: str | None = None
    custom: dict[str, Any] | None = None
    uncompressed_size: conint(ge=0) | None = None


class Debug(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    panic_file: str | None = Field(default=None, description="Path to the panic file if panic logging is enabled")


class DetailedError(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    code: int | None = None
    errors: list[str] | None = None
    message: str | None = None


class ErrorDetail(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    location: str | None = Field(
        default=None,
        description="Where the error occurred, e.g. 'body.items[3].tags' or 'path.thing-id'",
    )
    message: str | None = Field(default=None, description="Error message text")
    value: Any | None = Field(default=None, description="The value at the given location")


class ErrorModel(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    detail: str | None = Field(
        default=None,
        description="A human-readable explanation specific to this occurrence of the problem.",
        examples=["Property foo is required but is missing."],
    )
    errors: list[ErrorDetail] | None = Field(default=None, description="Optional list of individual error details")
    instance: AnyUrl | None = Field(
        default=None,
        description="A URI reference that identifies the specific occurrence of the problem.",
        examples=["https://example.com/error-log/abc123"],
    )
    status: int | None = Field(default=None, description="HTTP status code", examples=[400])
    title: str | None = Field(
        default=None,
        description="A short, human-readable summary of the problem type. This value should not change between occurrences of the error.",
        examples=["Bad Request"],
    )
    type: AnyUrl | None = Field(
        default="about:blank",
        description="A URI reference to human-readable documentation for the error.",
        examples=["https://example.com/errors/example"],
    )


class Features(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    auth_types: list[str] | None = None
    compression_plugins: list[str] | None = None
    metadata_plugins: list[str] | None = None
    operation_plugins: list[str] | None = None
    permissions_plugins: list[str] | None = None
    storage_plugins: list[str] | None = None


class FileDownloadTokenResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    token: str | None = None


class ListBinariesResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    available: list[BinaryInfo] | None = None


class OperationIdResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    id: str | None = Field(
        default=None,
        description="ID of the operation",
        examples=["2diK2kCkpgeHAQSNthIZ1JYyPte"],
    )
    location: str | None = Field(
        default=None,
        description="Location to poll for operation status, provided in the body for convenience",
        examples=["/api/v1/operations/2diK2kCkpgeHAQSNthIZ1JYyPte"],
    )


class OperationsRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    ids: list[str] | None = Field(default=None, description="List of operation IDs to retrieve")


class PlainSrcDst(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    dst: str | None = None
    src: str | None = None


class RemoveMetadataRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    paths: list[str] | None = Field(default=None, description="Paths to remove metadata for")
    recursive: bool | None = Field(default=None, description="Whether to remove metadata recursively")


class SetMetadataRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    metadata: dict[str, DataAssignment] | None = Field(default=None, description="Per-path metadata to set")


class StatusResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    build_info: BuildInfo | None = Field(default=None, description="Information about the build")
    debug: Debug | None = Field(default=None, description="Debugging information")
    features: Features | None = Field(default=None, description="List of features and plugins available in the server")
    ready: bool | None = Field(default=None, description="Indicates if the server is ready to accept requests")
    time: str | None = Field(default=None, description="Current server time")


class StorageConfigResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    storage: list[Any] | None = None


class StoragePath(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    path: str | None = Field(default=None, examples=["my/path/to/data"])
    remote: str | None = "any"


class TokenRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    other: dict[str, Any] | None = None
    read: list[str] | None = None
    recursive: bool | None = None
    type: str | None = None
    write: list[str] | None = None


class UserInfo(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    account_id: str | None = None
    groups: list[str] | None = None
    id: str | None = None
    is_admin: bool | None = None
    roles: list[str] | None = None
    service_admin: bool | None = None
    username: str | None = None


class WhoamiResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    user: UserInfo | None = None


class WorkerConfig(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    auth: dict[str, Any] | None = None
    compression: dict[str, Any] | None = None
    metadata: dict[str, Any] | None = None
    permissions: dict[str, Any] | None = None
    storage: list[Any] | None = None
    user_id: str | None = None


class WorkerResetResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    message: str | None = None


class WorkerTokenResponseBody(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    expires_at: AwareDatetime | None = None
    token: str | None = None


class ResourceType(Enum):
    Document = "document"


class RoleType(Enum):
    Reader = "reader"
    Writer = "writer"
    Admin = "admin"


class SubjectType(Enum):
    User = "user"
    Group = "group"
    Any = "any"


class OperationState(Enum):
    Unknown = "unknown"
    Queued = "queued"
    Running = "running"
    Succeeded = "succeeded"
    Failed = "failed"


class CopyMetadataRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    recursive: bool | None = Field(default=None, description="Whether to copy metadata recursively")
    src_dst: list[PlainSrcDst] | None = Field(default=None, description="Sources and destinations for the copy")


class GetMetadataRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    operations: list[StoragePath] | None = Field(
        default=None, deprecated=True, description="Deprecated as of 2024.09.10"
    )
    paths: list[str] | None = Field(default=None, description="Paths to retrieve metadata for")


class MoveMetadataRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    recursive: bool | None = Field(default=None, description="Whether to move metadata recursively")
    src_dst: list[PlainSrcDst] | None = Field(default=None, description="Sources and destinations for the move")


class Operation(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    children: list[str] | None = None
    children_detail: list[Operation] | None = None
    description: str | None = None
    ended_at: AwareDatetime | None = None
    error: str | None = None
    error_detail: DetailedError | None = None
    id: str | None = None
    info: dict[str, Any] | None = None
    labels: list[str] | None = None
    max_retries: int | None = None
    messages: list[str] | None = None
    progress: float | None = None
    progress_current: conint(ge=0) | None = None
    progress_total: conint(ge=0) | None = None
    queued_at: AwareDatetime | None = None
    result: Any | None = None
    retries: int | None = None
    started_at: AwareDatetime | None = None
    state: OperationState | None = None
    user_id: str | None = None


class OperationsResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    operations: list[Operation] | None = None


class PathOperations(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    operations: list[StoragePath] | None = None


class Resource(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    path: str | None = Field(default=None, examples=["my/path/to/data/file.txt"])
    type: ResourceType | None = Field(default=None, examples=["document"])


class SrcDst(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    dst: StoragePath | None = None
    src: StoragePath | None = None


class SrcDstOperations(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    operations: list[SrcDst] | None = None


class Subject(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    id: str | None = Field(default=None, examples=["946991ec-828c-4de4-acbe-962ada8bc441"])
    type: SubjectType | None = Field(default=None, examples=["user"])


class RoleAssignment(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    resource: Resource | None = None
    role: RoleType | None = None
    subject: Subject | None = None


class RoleQuery(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    resource: Resource | None = None
    role: RoleType | None = None
    subject: Subject | None = None


class SetPermissionsRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    permissions: list[RoleAssignment] | None = Field(default=None, description="Permission definitions to set")


class CheckPermissionsRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    permissions: list[RoleAssignment] | None = Field(default=None, description="Permission to check for")


class GetPermissionsRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    permissions: list[RoleQuery] | None = Field(default=None, description="Permission query to perform")


class GetPermissionsResponse(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    permissions: list[RoleAssignment] | None = Field(default=None, description="Permissions retrieved from the system")


class RemovePermissionsRequest(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    permissions: list[RoleAssignment] | None = Field(default=None, description="Permission to remove")


Operation.model_rebuild()



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for implementing retry mechanisms in the Ansys HPS data transfer client.

This module includes functionality for handling transient errors and ensuring
robust and reliable operations during data transfer.
"""

import logging
import os

import backoff
import httpx

from ..exceptions import HPSError, NotReadyError, TimeoutError

log = logging.getLogger(__name__)

max_tries_env_name = "ANSYS_DT_CLIENT_RETRY_MAX_TIME"
max_time_env_name = "ANSYS_DT_CLIENT_RETRY_MAX_TRIES"


def _on_backoff(details):
    try:
        msg = "Backing off {wait:0.1f} seconds after {tries} tries: {exception}".format(**details)
        log.info(msg)
    except Exception as ex:
        log.warning(f"Failed to log in backoff handler: {ex}")


def _giveup(e):
    if isinstance(e, httpx.ConnectError):
        return False
    elif isinstance(e, TimeoutError):
        return True
    elif isinstance(e, NotReadyError):
        return False
    elif isinstance(e, HPSError) and e.give_up:
        return True
    elif isinstance(e, TypeError):
        return True

    return False


def _lookup_max_time():
    v = os.getenv(max_time_env_name)
    if v is not None:
        return v
    return 300


def _lookup_max_tries():
    v = os.getenv(max_tries_env_name)
    if v is not None:
        return v
    return 40


def retry(
    max_tries=_lookup_max_tries,
    max_time=_lookup_max_time,
    raise_on_giveup=True,
    jitter=backoff.full_jitter,
):
    """Provides a decorator for retrying a function call with exponential backoff."""
    return backoff.on_exception(
        backoff.expo,
        Exception,
        max_tries=max_tries,
        max_time=max_time,
        jitter=jitter,
        raise_on_giveup=raise_on_giveup,
        on_backoff=_on_backoff,
        logger=None,
        giveup=_giveup,
    )



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides authentication for the user with a password or refresh token.

This module interacts with the HPS authentication service.
"""

import logging
import urllib.parse

import requests

from .exceptions import raise_for_status

log = logging.getLogger(__name__)


def authenticate(
    url: str = "https://localhost:8443/hps",
    realm: str = "rep",
    grant_type: str = "password",
    scope="openid",
    client_id: str = "rep-cli",
    client_secret: str = None,
    username: str = None,
    password: str = None,
    refresh_token: str = None,
    timeout: float = 10.0,
    verify: bool | str = True,
    **kwargs,
):
    """Authenticate the user with a password or refresh token against the HPS authentication service.

    If this method is successful, the response includes access and refresh tokens.

    Parameters
    ----------

    url : str, default: 'https://localhost:8443/hps'
        Base path for the server to call.
    realm : str, default: 'rep'
        Name of the Keycloak realm.
    grant_type: str, default: 'password'
        Authentication method.
    scope : str, default: 'openid'
        String containing one or more requested scopes.
    client_id : str, default: 'rep-cli'
        Client type.
    client_secret : str, default: None
        Client secret.
    username : str
        Username.
    password : str
        Password.
    refresh_token : str
        Refresh token.
    timeout : float, default: 10.0
        Timeout in seconds.
    verify: Union[bool, str]
        If a Boolean, whether to verify the server's TLS certificate. If a string, the
        path to the CA bundle to use. For more information, see the :class:`requests.Session`
        documentation.
    """
    auth_postfix = f"auth/realms/{realm}"
    if url.endswith(f"/{auth_postfix}") or url.endswith(f"/{auth_postfix}/"):
        auth_url = url
    else:
        auth_url = urllib.parse.urljoin(url + "/", auth_postfix)
    log.debug(f"Authenticating using {auth_url}")

    session = requests.Session()
    session.verify = verify
    session.headers = ({"content-type": "application/x-www-form-urlencoded"},)

    token_url = f"{auth_url}/protocol/openid-connect/token"

    data = {
        "client_id": client_id,
        "grant_type": grant_type,
        "scope": scope,
    }
    if client_secret is not None:
        data["client_secret"] = client_secret
    if username is not None:
        data["username"] = username
    if password is not None:
        data["password"] = password
    if refresh_token is not None:
        data["refresh_token"] = refresh_token

    data.update(**kwargs)

    log.debug(f"Retrieving access token for client {client_id} from {auth_url} using {grant_type} grant.")
    r = session.post(token_url, data=data, timeout=timeout)

    raise_for_status(r)
    return r.json()



# Copyright (C) 2025 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality for monitoring of operations."""

import datetime
import logging
import time

import humanize as hz

from ..models import Operation, OperationState

log = logging.getLogger(__name__)


class WaitHandler:
    """Allows additional handling of operation status on wait."""

    class Meta:
        """Meta class for WaitHandler."""

        expand_group = True

    final = [OperationState.Succeeded, OperationState.Failed]

    def __init__(self):
        """Initializes the WaitHandler class object."""
        self.start = time.time()
        self.report_threshold = 2.0  # seconds
        self.min_progress_interval = 15.0  # seconds
        self.last_progress = {}
        self.completed_ids = []

    def __call__(self, ops: list[Operation]):
        """Handle operations after they are fetched."""
        self._log_ops(ops)

    def _log_ops(self, ops: list[Operation]) -> str:
        # so_far = time.time() - self.start
        num_running = 0
        for op in ops:
            if op.children_detail is not None and self.Meta.expand_group:
                for ch in op.children_detail or []:
                    # For ops with lots of children, avoid logging completed children on every loop iteration
                    # If there are 29 small files and 1 large file in the group, we dont want to log 29 files
                    # are done 1000 times while the large file copies.
                    if ch.id not in self.completed_ids:
                        self._log_op(logging.DEBUG, ch)
                    if ch.state not in self.final:
                        num_running += 1
                    elif ch.id not in self.completed_ids:
                        self.completed_ids.append(ch.id)

            if op.state not in self.final:
                num_running += 1
            self._log_op(logging.INFO, op)

    def _log_op(self, lvl: int, op: Operation):
        """Format the operation description."""
        op_type = "operation" if op.children is None or len(op.children) == 0 else "operation group"
        op_done = op.state in self.final

        msg = f"Data transfer {op_type} '{op.description}'({op.id}) {'finished. ' if op_done else 'is in progress. '}"

        try:
            start = op.started_at
            if op_done:
                end = op.ended_at
            else:
                end = datetime.datetime.now(start.tzinfo)
            duration = (end - start).seconds
            duration_str = hz.precisedelta(end - start)
        except Exception as ex:
            log.debug(f"Failed to parse operation duration: {ex}")
            duration = 0
            duration_str = "unknown"

        # Initialize last progress time if not set, set it back in time
        # so it logs right away (report threshold) the first time.
        if self.last_progress.get(op.id, None) is None:
            self.last_progress[op.id] = time.time() - self.min_progress_interval

        state = op.state.value
        if op_done:
            msg += f"{state.title()} after {duration_str}"
            msg += self._info_str(op)
            # if op.messages:
            # msg += f', messages="{"; ".join(op.messages)}"'
            log.log(lvl, msg)
        elif duration > self.report_threshold and time.time() - self.last_progress[op.id] > self.min_progress_interval:
            self.last_progress[op.id] = time.time()
            msg += f"{state.title()} for {duration_str}"
            if op.progress and op.progress > 0.0:
                msg += f", progress {op.progress * 100.0:.1f}%"
            msg += self._info_str(op)
            log.log(lvl, msg)

    def _info_str(self, op: Operation) -> str:
        """Format the operation info."""
        if not op.info:
            return ""
        info = ", ".join([f"{k}={v}" for k, v in op.info.items()])
        return f", {info}"


class AsyncWaitHandler(WaitHandler):
    """Allows additional, asynchronous handling of operation status on wait."""

    def __init__(self):
        """Initializes the AsyncOperationHandler class object."""
        super().__init__()

    async def __call__(self, ops: list[Operation]):
        """Handle operations after they are fetched."""
        self._log_ops(ops)



# Copyright (C) 2022 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides asynchronous API functionality for interacting with the Ansys HPS data transfer client.

This module includes methods and utilities for performing
data transfer operations asynchronously, managing resources, and handling client interactions.
"""

import asyncio
import builtins
from collections.abc import Awaitable, Callable
import logging
import textwrap
import time
import traceback

import backoff
import humanize as hz

from ..client import AsyncClient
from ..exceptions import TimeoutError
from ..models import (
    CheckPermissionsResponse,
    DataAssignment,
    GetPermissionsResponse,
    Operation,
    OperationIdResponse,
    OperationsResponse,
    OperationState,
    RoleAssignment,
    RoleQuery,
    SetMetadataRequest,
    SrcDst,
    StatusResponse,
    StorageConfigResponse,
    StoragePath,
)
from ..utils.jitter import get_expo_backoff
from .handler import AsyncWaitHandler
from .retry import retry

log = logging.getLogger(__name__)


class AsyncDataTransferApi:
    """Provides a wrapper for the Data Transfer Worker REST API, offering an async interface."""

    def __init__(self, client: AsyncClient):
        """Initialize the async data transfer API with the client object."""
        self.dump_mode = "json"
        self.client = client
        self.wait_handler_factory = AsyncWaitHandler

    @retry()
    async def status(self, wait=False, sleep=5, jitter=True, timeout: float | None = 20.0):
        """Provides an async interface to get the status of the worker."""

        async def _sleep():
            log.info(f"Waiting for the worker to be ready on port {self.client.binary_config.port} ...")
            s = backoff.full_jitter(sleep) if jitter else sleep

            await asyncio.sleep(s)

        url = "/"
        start = time.time()
        while True:
            if timeout is not None and (time.time() - start) > timeout:
                raise TimeoutError("Timeout waiting for worker to be ready")

            resp = await self.client.session.get(url)
            json = resp.json()
            s = StatusResponse(**json)
            if wait and not s.ready:
                await _sleep()
                continue
            return s

    @retry()
    async def operations(self, ids: list[str], expand: bool = False):
        """Provides an async interface to get a list of operations by their IDs."""
        return await self._operations(ids, expand=expand)

    async def storages(self):
        """Provides an async interface to get the list of storage configurations."""
        url = "/storage"
        resp = await self.client.session.get(url)
        json = resp.json()
        return StorageConfigResponse(**json).storage

    async def copy(self, operations: list[SrcDst]):
        """Provides an async interface to copy a list of ``SrcDst`` objects."""
        return await self._exec_async_operation_req("copy", operations)

    async def exists(self, operations: list[StoragePath]):
        """Provides an async interface to check if a list of ``StoragePath`` objects exist."""
        return await self._exec_async_operation_req("exists", operations)

    async def list(self, operations: list[StoragePath]):
        """Provides an async interface to get a list of ``StoragePath`` objects."""
        return await self._exec_async_operation_req("list", operations)

    async def mkdir(self, operations: builtins.list[StoragePath]):
        """Provides an async interface to create a list of directories in the remote backend."""
        return await self._exec_async_operation_req("mkdir", operations)

    async def move(self, operations: builtins.list[SrcDst]):
        """Provides an async interface to move a list of ``SrcDst`` objects in the remote backend."""
        return await self._exec_async_operation_req("move", operations)

    async def remove(self, operations: builtins.list[StoragePath]):
        """Provides an async interface to remove files in the remote backend."""
        return await self._exec_async_operation_req("remove", operations)

    async def rmdir(self, operations: builtins.list[StoragePath]):
        """Provides an async interface to remove directories in the remote backend."""
        return await self._exec_async_operation_req("rmdir", operations)

    @retry()
    async def _exec_async_operation_req(
        self, storage_operation: str, operations: builtins.list[StoragePath] | builtins.list[SrcDst]
    ):
        url = f"/storage:{storage_operation}"
        payload = {"operations": [operation.model_dump(mode=self.dump_mode) for operation in operations]}
        resp = await self.client.session.post(url, json=payload)
        json = resp.json()
        return OperationIdResponse(**json)

    async def _operations(self, ids: builtins.list[str], expand: bool = False):
        url = "/operations"
        params = {"ids": ids}
        if expand:
            params["expand"] = "true"
        resp = await self.client.session.get(url, params=params)
        json = resp.json()
        return OperationsResponse(**json).operations

    @retry()
    async def check_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Provides an async interface to check permissions of a list of ``RoleAssignment`` objects."""
        url = "/permissions:check"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        resp = await self.client.session.post(url, json=payload)
        json = resp.json()
        return CheckPermissionsResponse(**json)

    @retry()
    async def get_permissions(self, permissions: builtins.list[RoleQuery]):
        """Provides an async interface to get permissions of a list of ``RoleQuery`` objects."""
        url = "/permissions:get"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        resp = await self.client.session.post(url, json=payload)
        json = resp.json()
        return GetPermissionsResponse(**json)

    @retry()
    async def remove_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Provides an async interface to remove permissions of a list of ``RoleAssignment`` objects."""
        url = "/permissions:remove"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        await self.client.session.post(url, json=payload)

    @retry()
    async def set_permissions(self, permissions: builtins.list[RoleAssignment]):
        """Provides an async interface to set permissions of a list of ``RoleAssignment`` objects."""
        url = "/permissions:set"
        payload = {"permissions": [permission.model_dump(mode=self.dump_mode) for permission in permissions]}
        await self.client.session.post(url, json=payload)

    @retry()
    async def get_metadata(self, paths: builtins.list[str | StoragePath]):
        """Provides an async interface to get metadata of a list of ``StoragePath`` objects."""
        url = "/metadata:get"
        paths = [p if isinstance(p, str) else p.path for p in paths]
        payload = {"paths": paths}
        resp = await self.client.session.post(url, json=payload)
        json = resp.json()
        return OperationIdResponse(**json)

    @retry()
    async def set_metadata(self, asgs: dict[str | StoragePath, DataAssignment]):
        """Provides an async interface to set metadata of a list of ``DataAssignment`` objects."""
        url = "/metadata:set"
        d = {k if isinstance(k, str) else k.path: v for k, v in asgs.items()}
        req = SetMetadataRequest(metadata=d)
        resp = await self.client.session.post(url, json=req.model_dump(mode=self.dump_mode))
        json = resp.json()
        return OperationIdResponse(**json)

    async def wait_for(
        self,
        operation_ids: builtins.list[str | Operation],
        timeout: float | None = None,
        interval: float = 0.1,
        cap: float = 2.0,
        raise_on_error: bool = False,
        handler: Callable[[builtins.list[Operation]], Awaitable[None]] = None,
    ):
        """Provides an async interface to wait for a list of operations to complete.

        Parameters
        ----------
        operation_ids: list[str | Operation]
            The list of operation ids to wait for.
        timeout: float | None
            The maximum time to wait for the operations to complete.
        interval: float
            The interval between checks for the operations to complete.
        cap: float
            The maximum backoff value used to calculate the next wait time. Default is 2.0.
        raise_on_error: bool
            Raise an exception if an error occurs. Default is False.
        operation_handler: Callable[[builtins.list[Operation]], None]
            A callable that will be called with the list of operations when they are fetched.
        """
        if handler is None:
            handler = self.wait_handler_factory()

        if not isinstance(operation_ids, list):
            operation_ids = [operation_ids]
        operation_ids = [op.id if isinstance(op, Operation | OperationIdResponse) else op for op in operation_ids]
        start = time.time()
        attempt = 0
        op_str = textwrap.wrap(", ".join(operation_ids), width=60, placeholder="...")
        # log.debug(f"Waiting for operations to complete: {op_str}")
        while True:
            attempt += 1
            try:
                expand = getattr(handler.Meta, "expand_group", False) if hasattr(handler, "Meta") else False
                ops = await self._operations(operation_ids, expand=expand)
                if handler is not None:
                    try:
                        await handler(ops)
                    except Exception as e:
                        log.warning(f"Handler error: {e}")
                        log.debug(traceback.format_exc())

                if all(op.state in [OperationState.Succeeded, OperationState.Failed] for op in ops):
                    break
            except Exception as e:
                log.debug(f"Error getting operations: {e}")
                if raise_on_error:
                    raise
                else:
                    break

            if timeout is not None and (time.time() - start) > timeout:
                raise TimeoutError("Timeout waiting for operations to complete")

            # TODO: Adjust based on transfer speed and file size
            duration = get_expo_backoff(interval, attempts=attempt, cap=cap, jitter=True)
            await asyncio.sleep(duration)

        duration = hz.naturalsize(time.time() - start)
        log.debug(f"Operations completed after {duration}: {op_str}")
        return ops
